/*
    _____ _       _   _        __    _     _   _   _            _____                     _
   |   __| |_ ___| |_|_|___   |  |  |_|___| |_| |_|_|___ ___   |   __|___ ___ ___ ___ ___| |_ ___ ___
   |__   |  _| .'|  _| |  _|  |  |__| | . |   |  _| |   | . |  |  |  | -_|   | -_|  _| .'|  _| . |  _|
   |_____|_| |__,|_| |_|___|  |_____|_|_  |_|_|_| |_|_|_|_  |  |_____|___|_|_|___|_| |__,|_| |___|_|
                                      |___|             |___|

       This generates a "static" light      ad888888b,
   map for your game. It can't update      d8"     "88   Copyright 2010 David "Darkcampainger" Braun
   at runtime, but it does generate an             a8P
   advanced, soft-shadow image.                 ,d8P"    Version 2.3 - May 30, 2011
                                              a8P"
   The 'Generate' verb will run the EXE     a8P'
   on the current map and display the      d8"
   results. The image is saved to          88888888888
   "lightMap_[name].dmi", which can be
   loaded later with the 'Load' verb

*/

//Layer shadow overlays should be place on
#define SHADOW_LAYER 10


/* Define the options for the generator:

     -a [ambientDarkness, 0-255 integer]       // Defines ambient level of darkness
     -f [falloff, 0-1 float]                   // Skew of light falloff
     -w [wallFade, 1-n float]                  // How much darker a wall is than its floor
     -p [perspective, 1-n integer]             // Scales down y-values by factor

     Example:
         Sets a high ambient darkness, and standard levels of falloff and wallFade:
             light_options = "-a 240 -f 0.4 -w 1.04"
*/
var/light_options = "-a 240 -f 0.4"

// We use our own opacity variable
atom/var/blocksLight=0
// Controls if surface acts like a vertical wall facing south
turf/var/lightCastWall=0

// If a turf has anything in it that blocks light, the entire tile blocks light
turf/proc/getOpacity()
	if(blocksLight) return 1
	for(var/atom/A in contents)
		if(A.blocksLight) return 1
	return 0

light
	parent_type=/obj
	var
		// Sets the center of the light (16.5 is the exact center)
		centerX=16.5;centerY=16.5
		// Sets how far it casts light (200 pixels)
		intensity=200
		// The radius of the source, determining how soft the shadows are (penumbra)
		radius=8

client/verb
	Generate()
		if(world.maxz<1)
			world.log<<"WARNING: No map included! A map must be loaded before generating! Try including one of the example maps in the \"Maps\" directory."
			return

		var/T = input("Enter file name. Extension and prefix will be added for you.","Generate Light Map: lightMap_\[name\].dmi") as text|null
		if(T)
			// Try to generate a lightMap for entire z-level 1
			if(light_generate("lightMap_[T].dmi",locate(1,1,1),locate(world.maxx,world.maxy,1)))
				// If successful, auto-load it at (1,1,1)
				light_load(file("lightMap_[T].dmi"), 1,1,1)

	Load(N as icon|null)
		if(N)
			// Load lightmap held in 'N' at (1,1,1)
			light_load(N,1,1,1)

proc
	// This process runs the generator on the area from A to B, and saves it to 'name'
	light_generate(name, turf/A, turf/B)

		if(istype(A,/atom/movable)) A=A.loc
		if(istype(B,/atom/movable)) B=B.loc
		if(!A || !B || !isturf(A) || !isturf(B)) {CRASH("bad min and max locations for generation area");}

		world<<"Encoding light map data into output file..."
		sleep(1)

		// Attempt to encode opacity and light data
		var/encodedMap = light_encode(A,B)
		if(!encodedMap) {CRASH("failed to encode map");}

		// Save the data to a text file
		if(fexists("lightMap_data.txt")) {fdel("lightMap_data.txt");}
		text2file(encodedMap,"lightMap_data.txt")

		world<<"Starting image generation... (this may take some time)"
		sleep(1)

		// Run the generator DLL
		var/r = light_runGenerator(name)

		// If generation failed, output the error message
		if(r != 0)
			if(r == null)
				world.log<<{"\nWARNING: If LightingGenerator.dll failed to load, ensure you have the 2010 C++ Runtime Libraries.
You can download the latest version from Microsoft: <a href="http://www.microsoft.com/downloads/en/details.aspx?FamilyID=a7b7a05e-6de6-4d3a-a423-37bf0912db84&amp;pf=true" target="_blank">(x86 32-bit System)</a> / <a href="http://www.microsoft.com/downloads/en/details.aspx?FamilyID=bd512d9e-43c8-4655-81bf-9350143d5867&amp;pf=true" target="_blank">(x64 64-bit System)</a>"}
				return 0
			else
				CRASH("generation failed in lightingGenerator.dll (code [r])")

		// Removed temporary file
		if(fexists("lightMap_data.txt")) {fdel("lightMap_data.txt");}

		world<<"Generation Complete! Light map compiled to: \n\t[name]"

		return 1

	// This is just a wrapper, in case call() crashes, so it doesn't crash light_generate()
	light_runGenerator(name)
		// Determine icon size
		var/iconWidth,iconHeight
		if(istext(world.icon_size))
			var/i = findtext(world.icon_size, "x")
			iconWidth = copytext(world.icon_size,1,i)
			iconHeight = copytext(world.icon_size,i+1)
		else
			iconWidth = world.icon_size
			iconHeight = world.icon_size

		// Run the generator with the supplied options
		return shell("LightingGenerator.exe [name] [light_options] -w [iconWidth] -h [iconHeight]")

	// Scans the map to generate a text string representation of the opacity
	// Encodes all light data
	light_encode(turf/Min,turf/Max)
		if(istype(Min,/atom/movable)) Min=Min.loc
		if(istype(Max,/atom/movable)) Max=Max.loc
		if(!Min || !Max || !isturf(Min) || !isturf(Max)) {CRASH("bad min and max locations for generation area");}

		var/list/lightArea = block(Min,Max)
		var/list/lights = list()
		var/encodedMap = "[Max.x-Min.x+1] [Max.y-Min.y+1]\n" //First two numbers in file are the width/height of map
		var/encodedWalls = "" // Encoded walls
		var/count=0,bool=1,B=0,wall=0,wallCount=0

		// Save opacity "chains". Similar to Run Length Encoding.
		for(var/turf/T in lightArea)

			// Only include lights in the designated area
			for(var/light/L in T.contents) {lights+=L}

			// Get the turf's light-opacity boolean
			B = T.getOpacity()

			if(B != bool) // Opacity chain broken, write the consecutive blocks up to now
				while(count >= 0)
					// Store number of consecutive blocks in ASCII value
					// ASCII 91 is a continuation char, meaning more of the *same* are coming
					encodedMap += ascii2text(min(91,count)+32)
					count-=91
				bool=B
				count=1
			else count++

			if(T.lightCastWall != wall) // Wall chain broken, write the consecutive blocks up to now
				while(wallCount >= 0)
					// Store number of consecutive blocks in ASCII value
					// ASCII 91 is a continuation char, meaning more of the *same* are coming
					encodedWalls += ascii2text(min(91,wallCount)+32)
					wallCount-=91
				wall=T.lightCastWall
				wallCount=1
			else wallCount++

		while(count >= 0) //Get final segment/chain
			encodedMap += ascii2text(min(91,count)+32)
			count-=91
		while(wallCount >= 0) //Get final segment/chain
			encodedWalls += ascii2text(min(91,wallCount)+32)
			wallCount-=91

		encodedMap += "\n[encodedWalls]"

		// Encode all the lights
		for(var/light/L in lights)
			encodedMap+="\n[L.x-Min.x] [L.y-Min.y] [L.centerX] [L.centerY] [L.intensity] [L.radius]"

		return encodedMap

	// Load previously generated light map, at (offsetX, offsetY, offsetZ)
	light_load(iconFile, offsetX=1, offsetY=1, offsetZ=1)
		if(iconFile)
			var/icon/I = icon(iconFile)
			var/turf/T

			// Go through the states, and parse their names to determine where to place light tile
			// State names in format "[x1],[y1],[x2],[y2],[...],[xn],[yn]"
			for(var/state in icon_states(I))
				var
					readIndex = 1
					findIndex = 1
					len = length(state)
					X;Y
				while(findIndex<=len && findIndex>0)
					findIndex = findtext(state, ",", readIndex)
					X = text2num(copytext(state,readIndex,findIndex))
					readIndex = findIndex+1

					findIndex = findtext(state, ",", readIndex)
					Y = text2num(copytext(state,readIndex,findIndex))
					readIndex = findIndex+1

					T = locate(X+offsetX,Y+offsetY,offsetZ)
					if(T) T.overlays+=image(icon=I,icon_state=state,layer=SHADOW_LAYER)



// --- Demo foo --- \\


world/view="23x23"
//world/icon_size = "32x16"

turf
	icon='turfs.dmi'
	ground/icon_state="white"
	wall {blocksLight=1;icon_state="wall"}
	sideWall {blocksLight=1;lightCastWall=1;icon_state="sideWall"}

light
	icon='turfs.dmi'
	icon_state="light"
	layer=SHADOW_LAYER+1
	Small
		intensity=100
		icon_state="slight"
		radius=4
	Huge
		intensity=400
		icon = 'hugeLight.dmi'
		radius = 48
		pixel_x = -32
		pixel_y = -32